from django.db.models.signals import post_delete
from django.dispatch import receiver

from elasticsearch import TransportError

from .models import Indexable


@receiver(post_delete)
def delete_es_index_on_delete(sender, instance, **kwargs):
    if isinstance(instance, Indexable):
        try:
            instance.delete_index(ignore=404)
        except TransportError as e:
            raise(e)



from django.conf import settings as _settings

ES_CONNECTIONS = {
    "default": {
        "hosts": "localhost"
    }
}
ES_INDEX = _settings.DATABASES["default"]["NAME"]
ES_INDEX_SETTINGS = {}

DJES_EXCLUDED_MODELS = []



from django.db import models
from django.db.models.fields.related import ManyToOneRel, ForeignObjectRel

from elasticsearch_dsl.mapping import Mapping
from elasticsearch_dsl.field import Field

from djes.conf import settings

FIELD_MAPPINGS = {
    "AutoField": {"type": "long"},
    "BigIntegerField": {"type": "long"},
    "BinaryField": {"type": "binary"},
    "BooleanField": {"type": "boolean"},
    "CharField": {"type": "string"},
    "CommaSeparatedIntegerField": {"type": "string"},
    "DateField": {"type": "date"},
    "DateTimeField": {"type": "date"},
    "DecimalField": {"type": "string"},
    "DurationField": {"type": "long"},
    "EmailField": {"type": "string"},
    # "FileField": {"type": ""},  # TODO: make a decision on this
    "FilePathField": {"type": "string"},
    "FloatField": {"type": "double"},
    # "ImageField": {"type": ""},  # TODO: make a decision on this
    "IntegerField": {"type": "long"},
    "IPAddressField": {"type": "string", "index": "not_analyzed"},
    "GenericIPAddressField": {"type": "string", "index": "not_analyzed"},
    "NullBooleanField": {"type": "boolean"},
    "PositiveIntegerField": {"type": "long"},
    "PositiveSmallIntegerField": {"type": "long"},
    "SlugField": {"type": "string", "index": "not_analyzed"},
    "SmallIntegerField": {"type": "long"},
    "TextField": {"type": "string"},
    "TimeField": {"type": "string"},
    "URLField": {"type": "string"},
    "UUIDField": {"type": "string", "index": "not_analyzed"},
    "ForeignKey": {"type": "long"},
    "ManyToManyField": {"type": "long"},
    "OneToOneField": {"type": "long"},
}


def get_first_mapping(cls):
    """This allows for Django-like inheritance of mapping configurations"""
    from .models import Indexable

    if issubclass(cls, Indexable) and hasattr(cls, "Mapping"):
        return cls.Mapping
    for base in cls.__bases__:
        mapping = get_first_mapping(base)
        if mapping:
            return mapping
    return None


class EmptyMeta(object):
    pass


class DjangoMapping(Mapping):
    """A subclass of the elasticsearch_dsl Mapping, allowing the automatic mapping
    of many fields on the model, while letting the developer override these settings"""

    def __init__(self, model):
        self.model = model
        if not hasattr(self, "Meta"):
            self.Meta = EmptyMeta

        default_name = "{}_{}".format(self.model._meta.app_label, self.model._meta.model_name)
        name = getattr(self.Meta, "doc_type", default_name)

        super(DjangoMapping, self).__init__(name)
        self._meta = {}

        excludes = getattr(self.Meta, "excludes", [])
        includes = getattr(self.Meta, "includes", [])

        for field in self.model._meta.get_fields():

            if field.auto_created and field.is_relation:
                if not hasattr(field, "rel") or not field.rel.parent_link:
                    continue

            manual_field_mapping = getattr(self, field.name, None)
            # TODO: I am 90% shirt this is not being utilized. Test later.
            if manual_field_mapping:
                self.field(field.name, manual_field_mapping)
                continue

            if field.name in excludes:
                continue

            self.configure_field(field)

        # Now any included relations
        for name in includes:
            field = self.model._meta.get_field(name)
            self.configure_field(field)

        # Now any custom fields
        for field in dir(self.__class__):
            manual_field_mapping = getattr(self, field)
            if field not in self.properties.properties.to_dict() and isinstance(manual_field_mapping, Field):
                self.field(field, manual_field_mapping)

        if getattr(self.Meta, "dynamic", "strict") == "strict":
            self.properties._params["dynamic"] = "strict"

    def configure_field(self, field):
        """This configures an Elasticsearch Mapping field, based on a Django model field"""
        from .models import Indexable

        # This is for reverse relations, which do not have a db column
        if field.auto_created and field.is_relation:
            if isinstance(field, (ForeignObjectRel, ManyToOneRel)) and issubclass(field.related_model, Indexable):
                related_properties = field.related_model.search_objects.mapping.properties.properties.to_dict()
                self.field(field.name, {"type": "nested", "properties": related_properties})
                return

        if field.get_internal_type() == "ManyToManyField" and issubclass(field.rel.to, Indexable):

            related_properties = field.rel.to.search_objects.mapping.properties.properties.to_dict()
            self.field(field.name, {"type": "nested", "properties": related_properties})
            return

        if isinstance(field, models.ForeignKey):
            # This is a related field, so it should maybe be nested?

            # We only want to nest fields when they are indexable, and not parent pointers.
            if issubclass(field.rel.to, Indexable) and not field.rel.parent_link:

                related_properties = field.rel.to.search_objects.mapping.properties.properties.to_dict()
                self.field(field.name, {"type": "nested", "properties": related_properties})
                return

        db_column, attname = field.db_column, field.attname

        field_args = FIELD_MAPPINGS.get(field.get_internal_type())
        if field_args:
            self.field(db_column or attname, field_args)
        else:
            raise Warning("Can't find {}".format(field.get_internal_type()))

    @property
    def index(self):
        return getattr(self.Meta, "index", settings.ES_INDEX)



from django.apps import AppConfig, apps

from .conf import settings

from elasticsearch_dsl.connections import connections


class IndexableRegistry(object):
    """Contains information about all PolymorphicIndexables in the project."""
    def __init__(self):
        self.all_models = {}
        self.families = {}
        self.indexes = {}

    def register(self, cls):
        """Adds a new PolymorphicIndexable to the registry."""
        doc_type = cls.search_objects.mapping.doc_type

        self.all_models[doc_type] = cls
        base_class = cls.get_base_class()
        if base_class not in self.families:
            self.families[base_class] = {}
        self.families[base_class][doc_type] = cls

        if cls.search_objects.mapping.index not in self.indexes:
            self.indexes[cls.search_objects.mapping.index] = []

        self.indexes[cls.search_objects.mapping.index].append(cls)


indexable_registry = IndexableRegistry()


class DJESConfig(AppConfig):
    name = 'djes'
    verbose_name = "DJ E.S."

    def ready(self):
        from .models import Indexable

        # Let's register all the Indexable models
        for model in apps.get_models():

            # If it quacks...
            if issubclass(model, Indexable):
                meta = getattr(model, "_meta")
                if meta and not getattr(meta, "abstract"):
                    indexable_registry.register(model)

    connections.configure(**settings.ES_CONNECTIONS)



import six

from djes.search import LazySearch
from .wrappers import SearchWrapper


class AugmentedSearchIterator(object):
    """This manages multiple LazySearch objects.self

     Allows for indexing multiple elasticsearch queries."""

    def __init__(self, *args):
        self._searches = self.init_searches(*args)
        self._results = []

    def __getitem__(self, n):
        if isinstance(n, int):
            if n >= len(self._results):
                _search = self.get_search(n)
                try:
                    self._results.append(six.next(_search))
                except StopIteration:
                    _search._exhausted = True
                    return self.__getitem__(n)
            return self._results[n]
        raise TypeError('''AugmentedSearchIterator can only be indexed by an integer''')

    def get_search(self, n):
        """Should return a search based on priority based on search exhuastion and validators.

        RULES:
        1. SearchWrapper.exhausted = False.
        2. First SearchWrapper.is_valid({result_index}) = True.
        3. If none of them return `is_valid({result_index}) = True` return first SearchWrapper.exhausted = False.
        """

        _search = None
        for search in self._searches:
            if search._exhausted:
                continue
            elif search.is_priority(n):
                _search = search
                break
            elif not _search:
                _search = search
        return _search

    def init_searches(self, *args):
        _searches = []
        for search in args:
            if isinstance(search, LazySearch):
                _searches.append(SearchWrapper(search))
            elif isinstance(search, tuple):
                if len(search) != 2:
                    raise ValueError(
                        "SearchWrapper tuple must be of length 2"
                        "e.g., (SearchInstance, [{validators..}])"
                    )
                _searches.append(SearchWrapper(search[0], validators=search[1]))
        return _searches

    @property
    def search(self):
        for search in self._searches:
            if not search._exhausted:
                return search



from django.db.backends import utils
from django.db import models
from django.db.models.fields.related import ManyToOneRel, ForeignObjectRel
from django.apps import apps

from elasticsearch_dsl import field

from six import iteritems


class ElasticSearchForeignKey(object):

    def __init__(self, field_name, model):
        self.model = shallow_class_factory(model)
        self.field_name = field_name

    def __get__(self, instance, objtype=None):
        return instance.__dict__.get(self.field_name)

    def __set__(self, instance, value):
        if isinstance(value, field.InnerObjectWrapper):
            value = value.to_dict()
        instance.__dict__[self.field_name] = self.model(**value)


class ElasticSearchRelatedManager(object):

    def __init__(self, instances):
        self.instances = instances

    def count(self):
        return len(self.instances)

    def all(self):
        return self.instances


class ElasticSearchManyField(object):

    def __init__(self, field_name, model):
        self.model = shallow_class_factory(model)
        self.field_name = field_name

    def __get__(self, instance, objtype=None):
        if instance is None:
            return property()
        return instance.__dict__.get(self.field_name, ElasticSearchRelatedManager([]))

    def __set__(self, instance, value):
        value = [item.to_dict() for item in value if isinstance(item, field.InnerObjectWrapper)]
        instance.__dict__[self.field_name] = ElasticSearchRelatedManager([self.model(**item_data) for item_data in value])


class Test:

    def get(self):
        pass

    attr = property(get, set)


def shallow_class_factory(model):
    if model._deferred:
        model = model._meta.proxy_for_model
    name = "{}_ElasticSearchResult".format(model.__name__)
    name = utils.truncate_name(name, 80, 32)
    # try to get the model from the django registry
    try:
        return apps.get_model(model._meta.app_label, name)
    # get the object's type - hopefully a django model
    except LookupError:
        class Meta(object):
            proxy = True
            app_label = model._meta.app_label

        class ElasticMapping:
            class Meta:
                elastic_abstract = True
                doc_type = model.search_objects.mapping.doc_type

        overrides = {
            "save": None,
            "Meta": Meta,
            "Mapping": ElasticMapping,
            "search_objects": model.search_objects.__class__(),
            "__module__": model.__module__,
            "_deferred": True,
        }

        for attname, es_field in iteritems(model.search_objects.mapping.properties._params["properties"]):
            if type(es_field) == field.Nested:
                # This is a nested object!
                dj_field = model._meta.get_field(attname)

                if isinstance(dj_field, ManyToOneRel):
                    overrides[attname] = ElasticSearchManyField(attname, dj_field.related_model)
                elif isinstance(dj_field, ForeignObjectRel):
                    overrides[attname] = ElasticSearchForeignKey(attname, dj_field.related_model)

                if isinstance(dj_field, models.ManyToManyField):
                    overrides[attname] = ElasticSearchManyField(attname, dj_field.rel.to)

                if isinstance(dj_field, models.ForeignKey):
                    # Let's add a fake foreignkey attribute
                    overrides[attname] = ElasticSearchForeignKey(attname, dj_field.rel.to)
        return type(str(name), (model,), overrides)



import six


class SearchWrapper(object):

        def __init__(self, search, validators=[]):
            self.search = search
            self._exhausted = False
            self._validators = self.check_validators(validators)

        def next(self):
            return six.next(self.search)

        def __next__(self):
            return self.next()

        def check_validators(self, validators):
            if not isinstance(validators, list):
                raise TypeError('''SearchWrapper validators must be passed in as a list.''')
            for vldtr in validators:
                if not callable(vldtr):
                        raise TypeError('''SearchWrapper validators must be callable.''')
            return validators

        def is_priority(self, *args):
            for validator in self._validators:
                vld = validator(*args)
                if not isinstance(vld, bool):
                    raise TypeError('''SearchWrapper validators must return a value of type `bool`''')
                elif vld:
                    return vld
            return False



import logging

from elasticsearch_dsl.connections import connections

from django.test import TestCase
from djes.management.commands.sync_es import get_indexes, sync_index


class BaseIndexableTestCase(TestCase):
    """A TestCase which handles setup and teardown of elasticsearch indexes."""

    elasticsearchLogger = logging.getLogger('elasticsearch')

    def setUp(self):
        """ If you're reading this. I am gone and dead, presumably.
            Elasticsearch's logging is quite loud and lets us know
            about anticipated errors, so I set the level to ERROR only.
            If elasticsearch is giving you trouble in tests and you
            aren't seeing any info, get rid of this. God bless you.
        """
        self.elasticsearchLogger.setLevel(logging.ERROR)
        self.es = connections.get_connection("default")
        self.indexes = get_indexes()

        for index in list(self.indexes):
            self.es.indices.delete_alias("{}*".format(index), "_all", ignore=[404])
            self.es.indices.delete("{}*".format(index), ignore=[404])

        for index, body in self.indexes.items():
            sync_index(index, body)

    def tearDown(self):
        for index in list(self.indexes):
            self.es.indices.delete_alias("{}*".format(index), "_all", ignore=[404])
            self.es.indices.delete("{}*".format(index), ignore=[404])



from django.db import models
from elasticsearch.exceptions import NotFoundError
from elasticsearch_dsl.connections import connections

from .apps import indexable_registry
from .factory import shallow_class_factory
from .mapping import DjangoMapping, get_first_mapping
from .search import LazySearch


class IndexableManager(models.Manager):
    """a custom manager class to handle integration of native django models and elasticsearch storage
    """

    search_class = LazySearch

    @property
    def client(self):
        """Get an elasticsearch client
        """
        if not hasattr(self, "_client"):
            self._client = connections.get_connection("default")
        return self._client

    @property
    def mapping(self):
        """Get a mapping class for this model

        This method will return a Mapping class for your model, generating it using settings from a
        `Mapping` class on your model (if one exists). The generated class is cached on the manager.
        """
        if not hasattr(self, "_mapping"):
            if hasattr(self.model, "Mapping"):
                mapping_klass = type("Mapping", (DjangoMapping, self.model.Mapping), {})
            else:
                mapping_klass = get_first_mapping(self.model)
                if mapping_klass is None:
                    mapping_klass = DjangoMapping
            self._mapping = mapping_klass(self.model)
        return self._mapping

    def from_es(self, hit):
        """Returns a Django model instance, using a document from Elasticsearch"""
        doc = hit.copy()
        klass = shallow_class_factory(self.model)

        # We can pass in the entire source, except when we have a non-indexable many-to-many
        for field in self.model._meta.get_fields():

            if not field.auto_created and field.many_to_many:
                if not issubclass(field.rel.to, Indexable):
                    if field.name in doc["_source"]:
                        del doc["_source"][field.name]

            # if field.one_to_many:
            #     if field.name in doc["_source"]:
            #         del doc["_source"][field.name]

        # Now let's go ahead and parse all the fields
        fields = self.mapping.properties.properties
        for key in fields:
            # TODO: What if we've mapped the property to a different name? Will we allow that?
            field = fields[key]

            # if isinstance(field, InnerObject):
            #     import pdb; pdb.set_trace()
            #     continue

            if doc["_source"].get(key):
                attribute_value = doc["_source"][key]

                doc["_source"][key] = field.to_python(attribute_value)

        return klass(**doc["_source"])

    def get(self, **kwargs):
        """Get a object from Elasticsearch by id
        """
        # get the doc id
        id = None
        if "id" in kwargs:
            id = kwargs["id"]
            del kwargs["id"]
        elif "pk" in kwargs:
            id = kwargs["pk"]
            del kwargs["pk"]
        else:
            raise self.model.DoesNotExist("You must provide an id to find")

        # connect to es and retrieve the document
        es = connections.get_connection("default")

        doc_type = self.model.search_objects.mapping.doc_type
        index = self.model.search_objects.mapping.index
        try:
            doc = es.get(index=index, doc_type=doc_type, id=id, **kwargs)
        except NotFoundError:
            message = "Can't find a document for {}, using id {}".format(
                doc_type, id)
            raise self.model.DoesNotExist(message)

        # parse and return
        return self.from_es(doc)

    def search(self):
        """
        """
        model_callbacks = {}
        indexes = []

        if self.model in indexable_registry.families:
            # There are child models...
            for doc_type, cls in indexable_registry.families[self.model].items():

                model_callbacks[doc_type] = cls.search_objects.from_es
                if cls.search_objects.mapping.index not in indexes:
                    indexes.append(cls.search_objects.mapping.index)
        else:
            # Just this one!
            model_callbacks[self.model.search_objects.mapping.doc_type] = self.from_es
            indexes.append(self.model.search_objects.mapping.index)

        return self.search_class().using(self.client).index(*indexes).doc_type(
            **model_callbacks
        )

    def refresh(self):
        """Force a refresh of the Elasticsearch index
        """
        self.client.indices.refresh(index=self.model.search_objects.mapping.index)


class Indexable(models.Model):
    """An abstract model that allows indexing and searching features
    """

    class Meta(object):
        abstract = True

    objects = models.Manager()
    search_objects = IndexableManager()

    def save(self, index=True, *args, **kwargs):
        super(Indexable, self).save(*args, **kwargs)
        if index:
            self.index()
        else:
            self.delete_index(ignore=[404])

    def to_dict(self):
        """Get a dictionary representation of this item, formatted for Elasticsearch"""
        out = {}

        fields = self.__class__.search_objects.mapping.properties.properties

        for key in fields:
            # TODO: What if we've mapped the property to a different name? Will we allow that?

            attribute = getattr(self, key)

            field = fields[key]

            # I believe this should take the highest priority.
            if hasattr(field, "to_es"):
                out[key] = field.to_es(attribute)

            # First we check it this is a manager, in which case we have many related objects
            elif isinstance(attribute, models.Manager):
                if issubclass(attribute.model, Indexable):
                    # TODO: We want this to have some awareness of the relevant field.
                    out[key] = [obj.to_dict() for obj in attribute.all()]
                else:
                    out[key] = list(attribute.values_list("pk", flat=True))

            elif callable(attribute):
                out[key] = attribute()

            elif isinstance(attribute, Indexable):
                out[key] = attribute.to_dict()
            else:
                out[key] = attribute

            if out[key] is None:
                del out[key]
        return out

    def index(self, refresh=False):
        """Indexes this object, using a document from `to_dict()`"""
        es = connections.get_connection("default")
        index = self.__class__.search_objects.mapping.index
        doc_type = self.__class__.search_objects.mapping.doc_type
        es.index(index, doc_type,
                 id=self.pk,
                 body=self.to_dict(),
                 refresh=refresh)

    def delete_index(self, refresh=False, ignore=None):
        """Removes the object from the index if `indexed=False`"""
        es = connections.get_connection("default")
        index = self.__class__.search_objects.mapping.index
        doc_type = self.__class__.search_objects.mapping.doc_type
        es.delete(index, doc_type, id=self.pk, refresh=refresh, ignore=ignore)

    @property
    def mapping(self):
        """Returns the proper mapping for this instance"""
        return self.__class__.search_objects.mapping

    @classmethod
    def is_orphaned(cls):
        manager = getattr(cls, 'search_objects', None)
        if not manager:
            return False
        mapping = getattr(manager, 'mapping', None)
        if not mapping:
            return False
        meta = getattr(mapping, 'Meta', None)
        if not meta:
            return False
        return getattr(meta, 'orphaned', False)

    @classmethod
    def is_elastic_abstract(cls):
        """HACK. In `factory.py we are not assigning Mapping the way we imagine to be.
        """
        return bool(
            cls.__name__.endswith("_ElasticSearchResult") or
            getattr(cls.search_objects.mapping, "elastic_abstract", False)
        )

    @classmethod
    def get_base_class(cls):
        if cls.is_orphaned():
            return cls
        if cls.__bases__:
            for base in cls.__bases__:
                if base == Indexable:
                    return cls
                elif hasattr(base, "get_base_class"):
                    base_base = base.get_base_class()
                    if base_base:
                        return base_base
        return None

    @classmethod
    def get_doc_types(cls, exclude_base=False):
        """Returns the doc_type of this class and all of its descendants."""
        names = []
        if not exclude_base and hasattr(cls, 'search_objects'):
            if not cls.is_elastic_abstract():
                names.append(cls.search_objects.mapping.doc_type)
        for subclass in cls.__subclasses__():
            names += subclass.get_doc_types()
        return names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.command.test import test as TestCommand  # noqa
import os
import re
import sys


name = 'djes'
package = 'djes'
description = "Connecting Django and elasticsearch-dsl"
url = "https://github.com/theonion/djes"
author = "Chris Sinchok"
author_email = 'csinchok@theonion.com'
license = 'MIT'

setup_requires = []

dev_requires = [
    "flake8==3.0.4",
    "pytest==2.8.7",
    "pytest-django==2.8.0",
    "pytest-cov==1.8.1",
    "model_mommy==1.2.4",
    "coveralls==0.5",
    "six==1.10.0",
    "ipdb",
]

install_requires = [
    "elasticsearch-dsl==0.0.9",
    "django>=1.8,<1.9"
]

server_requires = []

if 'test' in sys.argv:
    setup_requires.extend(dev_requires)


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("^__version__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['tests']
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(
    name=name,
    version=get_version(package),
    url=url,
    license=license,
    description=description,
    author=author,
    author_email=author_email,
    packages=get_packages(package),
    install_requires=install_requires,
    tests_require=dev_requires,
    extras_require={
        'dev': dev_requires,
    },
    cmdclass={'test': PyTest}
)
